// -*- C++ -*-
// $RCSfile: cmicrostructure.C,v $
// $Revision: 1.52 $
// $Author: reida $
// $Date: 2005/09/19 15:29:26 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <assert.h>
#include <map>
#include <vector>
#include <algorithm>
#include <stdlib.h>		// for abs()
#include <math.h>		// for sqrt()
#include <iomanip>

#include "common/cmicrostructure.h"
#include "common/coord.h"
#include "common/switchboard.h"
#include "common/geometry.h"
#include "common/printvec.h"
#include "common/pixelattribute.h"
#include "common/pixelsetboundary.h"

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CMicrostructure::CMicrostructure(const std::string &name,
				 const ICoord *isz, const Coord *sz) 
  : pxlsize_(*isz),
    size_(*sz),
    delta_((*sz)(0)/(*isz)(0), (*sz)(1)/(*isz)(1)),
    attributeMap(nAttributes()),
    attributeGlobalData(nAttributes()),
    categorymap((*isz)(0), (*isz)(1)),
    categorized(false),
    ncategories(0),
    name_(name)
{
  // Initialize pixel attribute maps
  for(std::vector<Array<PixelAttribute*> >::size_type i=0; i<attributeMap.size(); i++) {
    Array<PixelAttribute*> &map = attributeMap[i];
    const PxlAttributeRegistration *pareg =
      PxlAttributeRegistration::getRegistration(i);
    attributeGlobalData[i] = pareg->createAttributeGlobalData(this);
    map.resize(pxlsize_);
    for(Array<PixelAttribute*>::iterator j=map.begin(); j!=map.end(); ++j) {
      map[j] = pareg->createAttribute(j.coord());
    }
  }
}

CMicrostructure::~CMicrostructure() {
  destroy();
}

void CMicrostructure::destroy() {
  for(std::vector<PixelGroup*>::size_type i=0; i<pixelgroups.size(); i++) 
    delete pixelgroups[i];
  pixelgroups.resize(0);

  for(std::vector<Array<PixelAttribute*> >::size_type i=0; i<attributeMap.size(); i++) {
    Array<PixelAttribute*> &map = attributeMap[i];
    for(Array<PixelAttribute*>::iterator j=map.begin(); j!=map.end(); ++j)
      delete *j;
  }
  attributeMap.resize(0);
  for(std::vector<PixelAttributeGlobalData*>::size_type i=0; i<attributeGlobalData.size(); i++)
    delete attributeGlobalData[i];
  attributeGlobalData.resize(0);
}

Array<PixelAttribute*> &CMicrostructure::getAttributeMap(int which) const {
#ifdef DEBUG
  assert(which >= 0);
  assert(which < int(attributeMap.size()) );
#endif
  return attributeMap[which];
}

PixelAttributeGlobalData *CMicrostructure::getAttributeGlobalData(int which)
  const 
{
  return attributeGlobalData[which];
}

TimeStamp CMicrostructure::getTimeStamp() const {
  return timestamp;
}

int CMicrostructure::nGroups() const {
  return pixelgroups.size();
}

// Convert a Coord in the physical space to pixel coordinates (without
// rounding to the nearest integer).
Coord CMicrostructure::physical2Pixel(const Coord &pt) const {
  return Coord(pt(0)/delta_(0), pt(1)/delta_(1));
}

// Return the physical space coordinates of the lower-left corner of a
// pixel.
Coord CMicrostructure::pixel2Physical(const ICoord &pxl) const {
  return Coord(pxl(0)*delta_(0), pxl(1)*delta_(1));
}

// Return the physical space coordinates of a given non-integer
// coordinate in pixel space.
Coord CMicrostructure::pixel2Physical(const Coord &pt) const {
  return Coord(pt(0)*delta_(0), pt(1)*delta_(1));
}

// Return the coordinates of the pixel that contains the given point.
ICoord CMicrostructure::pixelFromPoint(const Coord &pt) const {
  Coord p = physical2Pixel(pt);
  double xx = p(0);
  double yy = p(1);
  if(xx == pxlsize_(0))
    xx -= 1.0;
  if(yy == pxlsize_(1))
    yy -= 1.0;
  return ICoord(int(xx), int(yy));
}

bool CMicrostructure::contains(const ICoord &ip) const {
  if ((ip(0)>=0 && ip(0)<pxlsize_(0)) && (ip(1)>=0 && ip(1)<pxlsize_(1)))
    return true;
  return false;
}

PixelGroup *CMicrostructure::findGroup(const std::string &name) const {
  // Get an existing group.  Don't create one if it doesn't already exist.
  for(std::vector<PixelGroup*>::size_type i=0; i<pixelgroups.size(); i++) {
    if(pixelgroups[i]->name() == name)
      return pixelgroups[i];
  }
  return 0;
}

PixelGroup *CMicrostructure::getGroup(const std::string &name) {
  // Get an existing group, or create one if necessary.
  PixelGroup *grp = findGroup(name);
  if(grp) return grp;
  PixelGroup *newgrp = new PixelGroup(name, &pxlsize_, this);
  pixelgroups.push_back(newgrp);
  ++timestamp;
  OOFMessage msg("new pixel group");
  msg.addarg(*newgrp);
  switchboard_notify(msg);
  return newgrp;
}

void CMicrostructure::removeGroup(const std::string &name) {
  // Remove a group from the list of pixel groups, and mark it as
  // defunct.  Defunct groups will be removed from the groupmap later.
  for(std::vector<PixelGroup*>::iterator i=pixelgroups.begin();
      i<pixelgroups.end(); ++i)
    {
      PixelGroup *grp = *i;
      if(grp->name() == name) {	// this is the group we're looking for
	pixelgroups.erase(i);	// remove from our list
 	grp->set_defunct();	// schedule for deletion
	defunctgroups.push_back(grp);
 	categorized = false;
 	++timestamp;

	OOFMessage msg("destroy pixel group");
	msg.addarg(*grp);
	msg.addarg(name_);
	switchboard_notify(msg);
	return;
      }
    }
}

std::vector<std::string> *CMicrostructure::groupNames() const {
  std::vector<std::string> *roster = new std::vector<std::string>;
  for(std::vector<PixelGroup*>::size_type i=0; i<pixelgroups.size(); ++i)
    roster->push_back(pixelgroups[i]->name());
  return roster;
}

// Comparison function for PixelAttribute vectors, used when mapping
// attributes to categories.

static bool ltAttributes(const std::vector<PixelAttribute*> &pavec0,
			 const std::vector<PixelAttribute*> &pavec1)
{
  int n0 = pavec0.size();
  int n1 = pavec1.size();
  int n = n0;
  if(n1 < n0) n = n1;
  for(int i=0; i<n; i++) {
    if(*pavec0[i] < *pavec1[i]) return true;
    if(*pavec1[i] < *pavec0[i]) return false;
  }
  if(n0 < n1) return true;
  return false;
}

// A CatMap is an STL map between vectors of PixelAttributes and
// categories.  This is not the same as CMicrostructure::categorymap,
// which is an array of categories (ie, a map from 2D space to
// categories).

typedef std::map<const std::vector<PixelAttribute*>, int,
		 bool (*)(const std::vector<PixelAttribute*>&,
			  const std::vector<PixelAttribute*>&) > CatMap;

void CMicrostructure::categorize() const {
  //  std::cerr << "CMicrostructure::categorize." << std::endl;
  CatMap catmap(ltAttributes);	// maps lists of groups to categories
  representativePixels.resize(0);
  categoryBdys.clear();  // PixelSetBoundary has no default constructor.
  ncategories = 0;
  int nattrs = attributeMap.size();
  // loop over pixels in the microstructure
  for(Array<int>::iterator i=categorymap.begin(); i!=categorymap.end(); ++i) {
    // construct a list of the attributes of this pixel
    std::vector<PixelAttribute*> attrs(nattrs);
    const ICoord &where = i.coord();
    for(int j=0; j<nattrs; j++)
      attrs[j] = attributeMap[j][where];
    // See if this list of attributes has been seen already
    CatMap::iterator cat = catmap.find(attrs);
    if(cat == catmap.end()) {	// category not found
      catmap[attrs] = ncategories;
      categorymap[i] = ncategories;
      // representativePixels.push_back(i.coord());
      representativePixels.push_back(where);

      PixelSetBoundary psb(this);
      psb.add_pixel(where);
      categoryBdys.push_back(psb);

      ncategories++;
    }
    else {
      categorymap[i] = (*cat).second; // assign previously found category
      categoryBdys[ (*cat).second ].add_pixel(where);
    }
  }

  // It may not be obvious to the casual reader, but at this point the
  // defunct pixel groups have been removed from the GroupLists, and
  // so the groups can actually be deleted.  What has happened is that
  // the search for categories has called operator<() on pairs of
  // PixelAttributes.  GroupList is a PixelAttribute.  GroupList's
  // operator<() calls GroupList.members(), which sorts the list and
  // removes the defunct groups from it.  PixelGroups are the only
  // PixelAttribute that is specifically stored in the Microstructure,
  // so they get special treatment here.

  if(defunctgroups.size() > 0) {
    for(std::vector<PixelGroup*>::size_type i=0; i<defunctgroups.size(); i++)
      delete defunctgroups[i];
    defunctgroups.resize(0);
  }

  // Find the boundaries.
  for(std::vector<PixelSetBoundary>::iterator i = categoryBdys.begin();
      i!=categoryBdys.end(); ++i ) {
    (*i).find_boundary();
  }

  categorized = true;
}

int CMicrostructure::nCategories() const {
  if(!categorized)
    categorize();
  return ncategories;
}

int CMicrostructure::category(const ICoord *where) const {
  if(!categorized)
    categorize();
  int cat = categorymap[*where];
  return cat;
}

int CMicrostructure::category(const ICoord &where) const {
  if(!categorized)
    categorize();
  return categorymap[where];
}

// Special version for finding the category of the pixel under an
// arbitrary point.
int CMicrostructure::category(const Coord &where) const {
  if(!categorized) 
    categorize();
  return categorymap[pixelFromPoint(where)];
}

int CMicrostructure::category(int x, int y) const {
  if(!categorized)
    categorize();
  return categorymap[ICoord(x,y)];
}

const ICoord &CMicrostructure::getRepresentativePixel(int category) const {
  if(!categorized)
    categorize();
  return representativePixels[category];
}

const Array<int> *CMicrostructure::getCategoryMap() const {
  if(!categorized)
    categorize();
  return &categorymap;
}

void CMicrostructure::recategorize() {
  categorized = false;
  ++timestamp;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Geometry routines for identifying pixels in the microstructure that
// are under segments and elements. 

// Return a list (vector) of pixels underlying a segment.  It's the
// responsibility of the caller to delete the vector.

std::vector<ICoord> *CMicrostructure::segmentPixels(const CSegment *seg) const {
  return segmentPixels(seg->p0, seg->p1);
}

std::vector<ICoord> *CMicrostructure::segmentPixels(const Coord &c0,
						    const Coord &c1) const
{
  // Coordinates of endpoints in pixel space (real).
  Coord p0(physical2Pixel(c0));
  Coord p1(physical2Pixel(c1));
  // Coordinates of pixels containing the endpoints (integer).
  ICoord ip0((int) floor(p0(0)), (int) floor(p0(1)));
  ICoord ip1((int) floor(p1(0)), (int) floor(p1(1)));
  // If an endpoint lies exactly on a pixel boundary, then the correct
  // choice for the pixel "containing" the endpoint depends on which
  // direction the segment is going.  The rule is that some part of
  // the segment must lie within the selected pixel.
  if(ip0(0) == p0(0) && ip1(0) < ip0(0))
    ip0(0) -= 1;
  if(ip1(0) == p1(0) && ip0(0) < ip1(0))
    ip1(0) -= 1;
  if(ip0(1) == p0(1) && ip1(1) < ip0(1))
    ip0(1) -= 1;
  if(ip1(1) == p1(1) && ip0(1) < ip1(1))
    ip1(1) -= 1;

  // For vertical and horizontal segments that exactly lie along
  // the pixel boundaries, we need to pick a right row or column of pixels.
  // For instance,
  //
  //  |    element B  |
  //  |xxxxxxxxxxxxxxx|
  //  b---------------a   segment a-b lies on the pixel boundary
  //  |ooooooooooooooo|
  //  |    element A  |
  //
  //  From element A's point of view, pixels along the segment a-b
  //  should be "ooooo", whereas from element B's p.o.v, corresponding
  //  pixels should be "xxxxx".
  //  Followings will deal with this adjustment.

  // Horizontal
  if ((p0(1) == p1(1) && p0(1) == ip0(1)) && p0(0) > p1(0)) {
    ip0(1) -= 1;
    ip1(1) -= 1;
  }
  // Vertical
  if ((p0(0) == p1(0) && p0(0) == ip0(0)) && p0(1) < p1(1)) {
    ip0(0) -= 1;
    ip1(0) -= 1;
  }

  // For vertical and horizontal segments on skeleton(microstructure) edges,
  // all that matters is
  // that the chosen pixels aren't outside the microstructure.  After
  // the preceding check, the only way that a end pixel can be outside
  // is if the segment lies along the top or right edges.
  int maxx = pxlsize_(0);
  int maxy = pxlsize_(1);
  if(ip0(0) == maxx && ip1(0) == maxx) {
    ip0(0) = maxx - 1;
    ip1(0) = maxx - 1;
  }
  if(ip0(1) == maxy && ip1(1) == maxy) {
    ip0(1) = maxy - 1;
    ip1(1) = maxy - 1;
  }

  ICoord id = ip1 - ip0;	// distance between pixel endpoints

  if(id(0) == 0 && id(1) == 0) { // segment is entirely within one pixel
    return new std::vector<ICoord>(1, ip0);
  }
  
  if(id(0) == 0) {	
    // Segment is contained within a single column of pixels.
    int npix = abs(id(1)) + 1;
    std::vector<ICoord> *pixels = new std::vector<ICoord>(npix);
    int x = ip0(0);
    int y0 = (ip0(1) < ip1(1)) ? ip0(1) : ip1(1);
    for(int i=0; i<npix; i++)
      (*pixels)[i] = ICoord(x, y0+i);
    return pixels;
  }
  
  if(id(1) == 0) {	
    // Segment is contained within a single row of pixels.
    int npix = abs(id(0)) + 1;
    std::vector<ICoord> *pixels = new std::vector<ICoord>(npix);
    int y = ip0(1);
    int x0 = (ip0(0) < ip1(0)) ? ip0(0) : ip1(0);
    for(int i=0; i<npix; i++)
      (*pixels)[i] = ICoord(x0+i, y);
    return pixels;
  }

  // segment is diagonal
  if(abs(id(0)) >= abs(id(1))) {
    // x range is bigger than y range, so loop over x.
    // Make sure that p0 is to the left of p1.
    if(ip0(0) > ip1(0)) {
      ICoord itemp(ip1);
      ip1 = ip0;
      ip0 = itemp;
      Coord temp(p1);
      p1 = p0;
      p0 = temp;
    }
    std::vector<ICoord> *pixels = new std::vector<ICoord>;
    pixels->reserve(2*abs(id(0))); // biggest possible size
    double x0 = p0(0);
    double y0 = p0(1);
    double slope = (p1(1) - y0)/(p1(0) - x0);
    pixels->push_back(ip0);
    // Iterate over columns of pixels.  Compute the y intercepts at
    // the boundaries between the columns (ie, integer values of x).
    // Whenever the integer part of the y intercept changes, we need
    // to include an extra pixel (at the new y value) in the previous
    // column.
    int lasty = ip0(1);	// previous integer part of y intercept
    for(int x=ip0(0)+1; x<=ip1(0); ++x) {
      int y = int(floor(y0 + slope*(x-x0)));
      if(y >=0 && y < maxy) {
	if(y != lasty) {
	  pixels->push_back(ICoord(x-1, y));
	}
	pixels->push_back(ICoord(x,y));
      }
      lasty = y;
    }
    // The last pixel may not have been included yet, if there's one
    // more y intercept change to come, so make sure it's included.
    if(pixels->back() != ip1)
      pixels->push_back(ip1);
    return pixels;
  } // abs(id(0)) >= abs(id(1))
  
  // abs(id(1)) > abs(id(0))
  // y range is bigger than x range, so loop over y.
  if(ip0(1) > ip1(1)) {
    ICoord itemp(ip1);
    ip1 = ip0;
    ip0 = itemp;
    Coord temp(p1);
    p1 = p0;
    p0 = temp;
  }
  std::vector<ICoord> *pixels = new std::vector<ICoord>;
  pixels->reserve(2*abs(id(1)));
  double x0 = p0(0);
  double y0 = p0(1);
  double slope = (p1(0) - x0)/(p1(1) - y0); // dx/dy
  pixels->push_back(ip0);
  int lastx = ip0(0);
  // iterate over rows of pixels
  for(int y=ip0(1)+1; y<=ip1(1); ++y) {
    int x = int(floor(x0 + slope*(y-y0)));
    if(x >= 0 && x < maxx) {
      if(x != lastx)
	pixels->push_back(ICoord(x, y-1));
      pixels->push_back(ICoord(x,y));
    }
    lastx = x;
  }
  if(pixels->back() != ip1)
    pixels->push_back(ip1);
  return pixels;
}

static const ICoord east(1, 0);
static const ICoord west(-1, 0);
static const ICoord north(0, 1);
static const ICoord south(0, -1);
static const ICoord northeast(1, 1);
static const ICoord northwest(-1, 1);
static const ICoord southeast(1, -1);
static const ICoord southwest(-1, -1);

MarkInfo::MarkInfo(const ICoord &size)
  : markedpixels(size, false)
{
  // Initial setting of markedregion will be overwritten... why is it done?
  markedregion = markedpixels.subarray(ICoord(0,0), markedpixels.size());
}


// Set the bounds of the marking region and clear it.  The
// markedpixels array is an array of booleans, used as an intermediate
// result when marking complicated things like elements.

MarkInfo *CMicrostructure::beginMarking(const CRectangle &bbox) const {
  MarkInfo *mm = new MarkInfo(sizeInPixels());
  ICoord p0 = pixelFromPoint(bbox.lowerleft());
  ICoord p1 = pixelFromPoint(bbox.upperright()) + northeast;
  if(p1(0) >= pxlsize_(0)) p1(0) = pxlsize_(0);
  if(p1(1) >= pxlsize_(1)) p1(1) = pxlsize_(1);
  mm->markedregion = mm->markedpixels.subarray(p0, p1);
  mm->markedregion.clear(false);
  return mm;
}

// Return a vector containing all the marked pixels.  It's the
// responsibility of the caller to delete the vector.

std::vector<ICoord> *CMicrostructure::markedPixels(MarkInfo *mm) const {
  return mm->markedregion.pixels(true); // returns new'd vector
}

// Mark the pixels underlying a segment.

void CMicrostructure::markSegment(MarkInfo *mm, 
				  const Coord &c0, const Coord &c1) const
{
  const std::vector<ICoord> *pixels = segmentPixels(c0, c1);
//   std::cerr << "CMicrostructure::markSegment: c0=" <<  c0 << " c1=" << c1 << std::endl;
  for(std::vector<ICoord>::const_iterator pxl=pixels->begin();
      pxl<pixels->end(); ++pxl)
    {
//       std::cerr << "CMicrostructure::markSegment: pxl=" << *pxl << std::endl;
      mm->markedregion.set(*pxl);
    }
//   std::cerr << "CMicrostructure:markSegment: deleting pixels" << std::endl;
  delete pixels;
//   std::cerr << "CMicrostructure::markSegment: done" << std::endl;
}


// Mark the pixels under a triangle by marking the pixels under its
// edges, then using a burn algorithm to mark the ones inside.

void CMicrostructure::markTriangle(MarkInfo *mm, const Coord &c0,
				   const Coord &c1, const Coord &c2)
  const
{
  markSegment(mm, c0, c1);
  markSegment(mm, c1, c2);
  markSegment(mm, c2, c0);
  // Find an unmarked point at which to start the burn algorithm.  For
  // small or narrow triangles, the center point may lie within a
  // pixel's distance of the edge and be already marked.
  ICoord start = pixelFromPoint(1./3.*(c0 + c1 + c2));
  if(mm->markedregion[start]) {
    // Center is already marked.  Unmarked pixels, if any, will lie
    // near the triangle's shortest edge.  Look for an unmarked pixel
    // by starting at the midpoint of the shortest edge and moving
    // towards the center of the triangle.  This code is really ugly
    // since it hardly seems worthwhile to put c0,c1,c2 in an array
    // and iterate over them.
    int shortest = 0;
    double shortlen = norm2(c1 - c2);
    double dd = norm2(c2 - c0);
    if(dd < shortlen) {
      shortlen = dd;
      shortest = 1;
    }
    dd = norm2(c0 - c1);
    if(dd < shortlen) {
      shortest = 2;
    }
    Coord mid;			// midpoint of shortest side
    Coord r;			// unit vector from center of short
				// side toward the opposite corner.
    switch (shortest) {
    case 0:
      mid = 0.5*(c1 + c2);
      r = c0 - mid;
      break;
    case 1:
      mid = 0.5*(c2 + c0);
      r = c1 - mid;
      break;
    case 2:
      mid = 0.5*(c0 + c1);
      r = c2 - mid;
      break;
    }
    double normr = sqrt(norm2(r));
    double maxdist = normr/3.0;	// no point in moving beyond center
    r = (1./normr)*r;		// unit vector
    double d = 1.0;
    bool ok = false;		// found a starting point yet?
    while(d < maxdist && !ok) {
      start = pixelFromPoint(mid + d*r);
      if(!mm->markedregion[start])
	ok = true;
      d += 1.0;
    }
    if(!ok) return;		// all interior pixels must already be marked
  }
  // Burn algorithm.  If a site is unmarked, the mark_site routine
  // marks it and puts it on the activesites list.  The main loop
  // (here) removes sites from the list and calls mark_site on their
  // neighbors.  Since the edges of the triangle are already marked,
  // they don't become active, and the burn won't spread beyond the
  // edges.
  std::vector<ICoord> activesites; // sites whose neighbors need to be checked
  mm->mark_site(activesites, start); // come on baby, light my fire
  while(activesites.size() > 0) {
    const ICoord here = activesites.back();
    activesites.pop_back();
    mm->mark_site(activesites, here + east);
    mm->mark_site(activesites, here + west);
    mm->mark_site(activesites, here + north);
    mm->mark_site(activesites, here + south);
    mm->mark_site(activesites, here + northeast);
    mm->mark_site(activesites, here + northwest);
    mm->mark_site(activesites, here + southeast);
    mm->mark_site(activesites, here + southwest);
  }
}

void MarkInfo::mark_site(std::vector<ICoord> &activesites,
			     const ICoord &here)
{
  if(markedregion.contains(here) && !markedregion[here]) {
    markedregion.set(here);
    activesites.push_back(here);
  }
}

void CMicrostructure::endMarking(MarkInfo *mm) const {
  delete mm;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Find the point on the line joining c0 and c1 at which the pixel
// category changes, and return it in result.  The return value is
// true if there's exactly one transition point.

bool CMicrostructure::transitionPoint(const Coord &c0, const Coord &c1,
				      Coord *result)
  const
{
  const std::vector<ICoord> *pixels = segmentPixels(c0, c1);
  if(pixels->size() < 2) {
    delete pixels;
    return false;
  }

  const Coord p0(physical2Pixel(c0));
  const Coord p1(physical2Pixel(c1));
  const Coord delta(p1 - p0);
  bool found = false;

  if(delta(0) == 0.0) {		// Infinite slope, vertical column of pixels
    std::vector<ICoord>::const_iterator pixel = pixels->begin();
    int prevcat = category(*pixel); // previous category
    ++pixel;
    for( ; pixel<pixels->end(); ++pixel) {
      int cat = category(*pixel);
      if(cat != prevcat) {	// found a transition
	if(found) {		// found too many of them
	  found = false;
	  break;
	}
	// segmentPixels() always goes from bottom to top, so we don't
	// need to worry about which edge of the pixel the transition
	// is on.
	*result = pixel2Physical(Coord(p0(0), (*pixel)(1)));
	found = true;
	prevcat = cat;
      }
    }
  }
  else {			// The slope isn't infinite. 
    double slope = (p1(1) - p0(1))/(p1(0) - p0(0));
    double invslope = 1./slope;
    double x0 = p0(0);
    double y0 = p0(1);
    std::vector<ICoord>::const_iterator pixel = pixels->begin();
    ICoord prevpxl = *pixel;	// previous pixel
    ++pixel;
    int prevcat = category(prevpxl); // previous category
    for( ; pixel<pixels->end(); ++pixel) {
      int cat = category(*pixel);
      if(cat != prevcat) {
	if(found) {
	  delete pixels;
	  return false;		// found too many transitions
	}
	// Find the transition point.  First decide which pixel
	// boundary it's on.  Consecutive pixels in the list returned
	// by segmentPixels() are always share a side (not just a
	// corner), so there are four cases to check.
	const ICoord diff(*pixel - prevpxl);
	double x, y;
	if(diff(0) == 1) {	// moving right, take right edge of last point
	  x = prevpxl(0) + 1.0;
	  y = y0 + (x-x0)*slope;
	}
	else if(diff(0) == -1) { // moving left, take left edge of last point
	  x = prevpxl(0);
	  y = y0 + (x-x0)*slope;
	}
	else if(diff(1) == -1) { // moving down, take bottom edge of last point
	  y = prevpxl(1);
	  x = x0 + (y-y0)*invslope;
	}
	else {			// moving up, take top edge of last point
	  y = prevpxl(1) + 1.0;
	  x = x0 + (y-y0)*invslope;
	}
	*result = pixel2Physical(Coord(x,y));
	found = true;
	prevcat = cat;
      }
      prevpxl = *pixel;
    }
  }
  delete pixels;
  return found;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double CMicrostructure::edgeHomogeneity(const Coord &c0, const Coord &c1) const
{
  const std::vector<ICoord> *pixels = segmentPixels(c0, c1);
  if(pixels->size() < 2) {
    delete pixels;
    return 1.0;
  }
  const Coord p0(physical2Pixel(c0));
  const Coord p1(physical2Pixel(c1));
  const Coord delta(p1 - p0);
  std::vector<double> lengths(nCategories(), 0.0);
  
  if(delta(0) == 0.0) {		// Infinite slope, vertical column of pixels
    // segmentPixels is guaranteed to return pixels in order of increasing y.
    std::vector<ICoord>::const_iterator pixel = pixels->begin();
    double ymin, ymax;
    if(p0(1) < p1(1)) {
      ymin = p0(1);
      ymax = p1(1);
    }
    else {
      ymin = p1(1);
      ymax = p0(1);
    }
    double prevy = ymin;
    int prevcat = category(*pixel);
    ++pixel;
    for( ; pixel<pixels->end(); ++pixel) {
      int cat = category(*pixel);
      if(cat != prevcat) {
	double y = (*pixel)(1);
	lengths[prevcat] += y - prevy;
	prevy = y;
	prevcat = cat;
      }
    }
    // The endpoint is within the last pixel, but we've only checked
    // as far as the bottom of the pixel...
    lengths[prevcat] += ymax - prevy;
  }

  else {			// Finite slope.
    double slope = delta(1)/delta(0);
    double invslope = 0.0;
    if(slope != 0.0)		// invslope won't be needed if slope==0.0
      invslope = 1./slope;
    double x0 = p0(0);
    double y0 = p0(1);
    Coord prevpt;		// previous point looked at
    Coord finalpt;		// last point to be looked at
    std::vector<ICoord>::const_iterator pixel = pixels->begin();
    ICoord prevpxl(*pixel);	// previous pixel
    int prevcat = category(prevpxl); // previous category
    if(prevpxl(0) <= p0(0) && prevpxl(0)+1 >= p0(0) &&
       prevpxl(1) <= p0(1) && prevpxl(1)+1 >= p0(1)) {
      prevpt = p0;
      finalpt = p1;
    }
    else {
      prevpt = p1;
      finalpt = p0;
    }
    ++pixel;
    for( ; pixel<pixels->end(); ++pixel) {
      int cat = category(*pixel);
      if(cat != prevcat) {
	// Find where the line entered this pixel.
	ICoord diff(*pixel - prevpxl);
	double x, y;
	if(diff(0) == 1) {	// moving right, use right edge of last pixel
	  x = prevpxl(0) + 1.0;
	  y = y0 + (x-x0)*slope;
	}
	else if(diff(0) == -1) { // moving left
	  x = prevpxl(0);
	  y = y0 + (x-x0)*slope;
	}
	else if(diff(1) == -1) { // moving down
	  y = prevpxl(1);
	  x = x0 + (y-y0)*invslope;
	}
	else {			// Mama, if that's movin' up then I'm ...
	  y = prevpxl(1) + 1.0;	// movin' out.
	  x = x0 + (y-y0)*invslope;
	}
	Coord point(x, y);
	lengths[prevcat] += sqrt(norm2(point - prevpt));
	prevpt = point;
	prevcat = cat;
      }
      prevpxl = *pixel;
    }
    lengths[prevcat] += sqrt(norm2(finalpt - prevpt));
  }

  delete pixels;

  double max = 0.0;
  for(std::vector<double>::size_type i=0; i<lengths.size(); i++)
    if(max < lengths[i])
      max = lengths[i];
  return max/sqrt(norm2(delta));
}


