# -*- python -*-
# $RCSfile: microstructure.py,v $
# $Revision: 1.124 $
# $Author: langer $
# $Date: 2005/12/18 00:27:01 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A Microstructure object is a segmentation of an Image.  It stores the
# Pixelgroups to which the pixels in the Image have been assigned, as
# well as other PixelAttributes, such as Materials.

from oof2.SWIG.common import activearea
from oof2.SWIG.common import cmicrostructure
from oof2.SWIG.common import pixelgroup
from oof2.SWIG.common import switchboard
from oof2.common import color
from oof2.common import debug
from oof2.common import parallel_enable
from oof2.common import pixelselection
from oof2.common import primitives
from oof2.common import utils
if parallel_enable.query() == 1:
    from oof2.common.IO import microstructureIPC
from oof2.common.IO import whoville
from types import *
import string

#############################

# MicrostructurePlugIns are objects that can be defined in other
# modules.  An instance of each plug-in is created when a
# Microstructure is created, and destroyed when it's destroyed.  What
# happens in between is up to the plug-in.

class MicrostructurePlugIn:
    def __init__(self, microstructure):
        self.microstructure = microstructure
    def destroy(self):
        self.microstructure = None

plugInClasses = []

def registerMicrostructurePlugIn(plugIn):
    plugInClasses.append(plugIn)

#############################


class Microstructure(cmicrostructure.CMicrostructure):
    def __init__(self, name, isize, size):
        cmicrostructure.CMicrostructure.__init__(self, name, isize, size)
        self._isize = isize             # iPoint(width, height) in pixels
        self._size = size               # Point(width, height) in physical units
        # Also store increments.
        self._delta  = (self._size[0]/(self._isize[0]*1.0),
                        self._size[1]/(self._isize[1]*1.0))

        self.images = []

        pixsel = pixelselection.PixelSelection(isize, size, self)
        pixelselection.pixelselectionWhoClass.add(name, pixsel)
        self.pixelselection = pixelselection.pixelselectionWhoClass[name]

        self.activearea = activearea.activeareaWhoClass.add(
            name, activearea.ActiveArea(isize, size, self))
        self.namedActiveAreas = []
        self.setCurrentActiveArea(self.activearea.getObject())

        # Add to list of all Microstructures.  Do this last, because
        # it calls switchboard callbacks that may assume that the MS
        # is fully constructed.
        microStructures.add(name, self)

        self.sbcallbacks = [
            switchboard.requestCallback(('who changed', 'Active Area'),
                                        self.aaChangedCB)
            ]
        self.plugins = []
        for pluginclass in plugInClasses:
            self.plugins.append(pluginclass(self))

    def aaChangedCB(self, activearea):
        if activearea.name() == self.name():
            self.setCurrentActiveArea(activearea.getObject())

    def nominalCopy(self, name):
        return Microstructure(name, self._isize, self._size)
        
    def destroy(self):
        for plugin in self.plugins:
            plugin.destroy()
        map(switchboard.removeCallback, self.sbcallbacks)
        for image in self.images:
            image.removeMicrostructure(self)
            whoville.getClass('Image').remove([self.name(), image.name()])
        self.images = []
        cmicrostructure.CMicrostructure.destroy(self)
        activearea.activeareaWhoClass.remove(self.name())
        pixelselection.pixelselectionWhoClass.remove(self.name())
        self.pixelselection = None
        self.activearea = None

    def addImage(self, image):
        # image is an ImageContext object, not an OOFImage.
        self.images.append(image)
        image.addMicrostructure(self)
        switchboard.notify('images changed in microstructure')
        if parallel_enable.query() == 1:
            try:
                from oof2.SWIG.common import mpitools
                if mpitools.rank() == 0:
                    microstructureIPC.ipcmicrostructuremenu.AddImageContext(microstructure = self.name(), imagecontext = image.name())
            except ImportError:
                pass
    def removeImage(self, image):
        self.images.remove(image)
        image.removeMicrostructure(self)
        whoville.getClass('Image').remove([self.name(), image.name()])
        switchboard.notify('images changed in microstructure')

    def imageNames(self):
        return [image.name() for image in self.images]

##    def name(self):
##        return self._name

##    def rename(self, name):
##        self._name = name


    # Many of these small utility functions are duplicated in C++ in
    # CMicrostructure.  That's intentional.
    
    def sizeInPixels(self):
        "iPoint(width, height) in pixels"
        return self._isize

    def sizeOfPixels(self):
        "(delta-x, delta-y) in a tuple."
        return self._delta

    def areaOfPixels(self):
        return self._delta[0]*self._delta[1]

    def area(self):
        return self._size[0]*self._size[1]
    
    def size(self):
        "Point(width, height) in physical units"
        return self._size

    def coords(self):
        ip = primitives.iPoint
        return [ip(i,j)
                for j in range(self._isize[1])
                for i in range(self._isize[0])]

    def renameGroup(self, oldname, newname):
        grp = self.findGroup(oldname)
        grp.rename(self.uniqueGroupName(newname, oldname))

    def uniqueGroupName(self, name, oldname=None):
        return utils.uniqueName(name, self.groupNames(), exclude=oldname)

    # For this one, "where" is a Point.
    def categoryFromPoint(self, where):
        xx=where.x/self._delta[0]
        if xx==self._isize[0]: xx=self._isize[0]-1
        yy=where.y/self._delta[1]
        if yy==self._isize[1]: yy=self._isize[1]-1
        return self.category(primitives.iPoint(xx,yy))

    # Return the iPoint pixel index corresponding to the passed-in Point.
    def pixelFromPoint(self, where):
        xx=where.x/self._delta[0]
        if xx==1.0*self._isize[0]: xx=self._isize[0]-1
        yy=where.y/self._delta[1]
        if yy==1.0*self._isize[1]: yy=self._isize[1]-1
        return primitives.iPoint(xx,yy)

    def pointFromPixel(self, where):
        # This returns the center of the pixel, given integer pixel
        # coordinates. This is not the same thing as the general
        # transformation from (floating point) pixel coordinates to
        # physical coordinates, which is done by
        # CMicrostructure::pixel2Physical.
        return primitives.Point((where.x+0.5)*self._delta[0],
                                (where.y+0.5)*self._delta[1])

    def activePoint(self,point):
        # Is the given point in the active area?  No need to check
        # bounds, since isSelected calls BitmapOverlay::get(), which
        # does the checking.
        return self.activearea.isActive(self.pixelFromPoint(point))

    def saveActiveArea(self, name):
        aa = self.activearea.getObject().clone()
        self.namedActiveAreas.append(activearea.NamedActiveArea(name, aa))
        for pxl in aa.members():
            activearea.areaListFromPixel(self, pxl).add(name)
        
    def renameActiveArea(self, oldname, newname):
        for naa in self.namedActiveAreas:
            if naa.name == oldname:
                real_new_name = utils.uniqueName(newname,
                                                 self.activeAreaNames(),
                                                 exclude=oldname)
                naa.name = real_new_name
                for pxl in naa.activearea.members():
                    aal = activearea.areaListFromPixel(self, pxl)
                    aal.remove(oldname)
                    aal.add(real_new_name)

    def deleteActiveArea(self, name):
        for i in range(len(self.namedActiveAreas)):
            if self.namedActiveAreas[i].name == name:
                aa = self.namedActiveAreas[i]
                for pxl in aa.activearea.members():
                    activearea.areaListFromPixel(self, pxl).remove(name)
                del self.namedActiveAreas[i]
                return
            
    def getNamedActiveArea(self, name):
        for naa in self.namedActiveAreas:
            if naa.name == name:
                return naa
    def activeAreaNames(self):
        return [aa.name for aa in self.namedActiveAreas]
        

    # Microstructures are fairly complex objects, hard to construct,
    # but easy to retrieve, so the repr is a "retriever".
    def __repr__(self):
        return "getMicrostructure(%s)" % `self.name()`


class MicrostructureContext(whoville.Who):
    def getMicrostructure(self):
        return self.getObject()
    def getSelectionContext(self):
        return self.getObject().pixelselection
    def parallelize(self):
        if parallel_enable.query() == 1:
            try:
                from oof2.SWIG.common import mpitools
                if mpitools.rank() == 0:
                    ## print self.name()
                    ms = self.getObject()
                    microstructureIPC.ipcmicrostructuremenu.Initialize(microstructure = self.name(), pixelsize = ms.sizeInPixels(), physicalsize = ms.size())
            except ImportError:
                pass
        return

## microSctructures is an instance of the WhoClass, which hosts
## a set of MicrostructureContext objects. The MicrostructureContext
## class was subclassed from the Who class, and it contains
## a Microstructure object.
## Please, also see note in image/oofimage.spy
microStructures = whoville.WhoClass('Microstructure', 150,
                                    instanceClass=MicrostructureContext,
                                    proxyClasses=['<topmost>'])

# Handy utility function, returns the microstructure object, not the
# context, and so is not equivalent to microStrutctures[name].
def getMicrostructure(name):
    try:
        return microStructures[name].getObject()
    except KeyError:
        return None

def getMSContextFromMS(microstructure):
    return microStructures[microstructure.name()]

utils.OOFdefine('Microstructure', Microstructure)
utils.OOFdefine('getMicrostructure', getMicrostructure)




