# -*- python -*-
# $RCSfile: fieldinit.py,v $
# $Revision: 1.32 $
# $Author: langer $
# $Date: 2005/12/22 21:56:54 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.engine import field
from oof2.common import debug
from oof2.common import registeredclass
from oof2.common import strfunction
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import xmlmenudump
import oof2.engine.mesh
import string
import struct
import types

# FieldInit subclasses must be derived from FieldInit *and* from
# RegisteredClass.

class FieldInit:
    def apply(self, femesh, field):
        fniter = femesh.funcnode_iterator()
        while not fniter.end():
            node = fniter.node()
            position = node.position()
            for i in range(field.ndof()):
                field.setvalue(femesh, node, i, self.func(position, i))
            fniter.next()

# The FieldInitParameter is not a simple RegisteredParameter because
# it has to handle more than one RegisteredClass, for different
# varieties of Fields. 

class FieldInitParameter(parameter.Parameter):
    types = (FieldInit,)
    def binaryRepr(self, datafile, value):
        return value.binaryRepr(datafile) # uses RegisteredClass.binaryRepr
    def binaryRead(self, parser):
        (regkey,) = struct.unpack('>i', parser.getBytes(struct.calcsize('>i')))
        registration = parser.getObject(regkey)
        argdict = {}
        for param in registration.params:
            argdict[param.name] = param.binaryRead(parser)
        return registration(**argdict)
    def valueDesc(self):
        return """A field initializer from one of the <link
        linkend='RegisteredClass:ScalarFieldInit'><classname>ScalarFieldInit</classname></link>
        or <link
        linkend='RegisteredClass:TwoVectorFieldInit'><classname>TwoVectorFieldInit</classname></link>
        classes."""

#############################

# Derived classes of FieldInit must provide func(position, component)
# that returns the value of a Field component at the given position.
# They must also indicate what type of field they work for by
# inserting themselves into the fieldInitDict dictionary.

fieldInitDict = {}

class ScalarFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []
    tip="Methods for initializing a scalar field."
    discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/scalar_field_init.xml')
fieldInitDict['ScalarField'] = ScalarFieldInit

class TwoVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
    registry = []
    tip="Methods for initializing a two dimensional vector field."
    discussion=xmlmenudump.loadFile(
        'DISCUSSIONS/engine/reg/two_vector_field_init.xml')
fieldInitDict['TwoVectorField'] = TwoVectorFieldInit

# Subclasses of this are stubs, existing mainly to populate the
# dictionary -- this class does everything.
class OtherMeshFieldInit:
    def __init__(self, mesh):
        self.mesh = mesh                # name of the other mesh
        self.meshobj = oof2.engine.mesh.meshes[self.mesh]
        self.source_femesh = self.meshobj.getObject()
    def apply(self, femesh, field):
        femesh.init_field(self.meshobj.skeleton, self.source_femesh, field)
    def shortrepr(self):
        return "mesh=%s" % self.mesh

# NVectorField isn't swigged yet
##class NVectorFieldInit(registeredclass.RegisteredClass, FieldInit):
##    registry = []
##fieldInitDict[field.NVectorField] = NVectorFieldInit
   


################    

class ConstScalarFieldInit(ScalarFieldInit):
    def __init__(self, value):
        self.value = value
    def func(self, position, component):
        return self.value
    def shortrepr(self):
        return str(self.value)

registeredclass.Registration(
    "Constant",
    ScalarFieldInit,
    ConstScalarFieldInit,
    0,
    params= [parameter.FloatParameter('value', 0.0,
                                      tip='Value to assign to the Field.')],
    tip="Initialize a scalar field with a constant value.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/const_scalar_field_init.xml')
    )

################

class ConstTwoVectorFieldInit(TwoVectorFieldInit):
    def __init__(self, cx, cy):
        self.cx = cx
        self.cy = cy
    def func(self, position, component):
        if component == 0:
            return self.cx
        return self.cy
    def shortrepr(self):
        return "cx=%s, cy=%s" % (str(self.cx), str(self.cy))

registeredclass.Registration(
    "Constant",
    TwoVectorFieldInit,
    ConstTwoVectorFieldInit,
    0,
    params=[
    parameter.FloatParameter('cx', 0.0,
                      tip='Value to assign to the x component of the Field.'),
    parameter.FloatParameter('cy', 0.0,
                      tip='Value to assign to the y component of the Field.')
    ],
    tip="Initialize a two-vector field with a constant value.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/const_twovec_init.xml')
    )

################

class FuncScalarFieldInit(ScalarFieldInit):
    def __init__(self, function):
        self.function = function
    def func(self, position, component):
        return self.function(position.x, position.y)
    def shortrepr(self):
        return str(self.function)

registeredclass.Registration(
    "XYFunction",
    ScalarFieldInit,
    FuncScalarFieldInit,
    1,
    params=[
    strfunction.XYStrFunctionParameter('function', default='0.0',
                                       tip='A function of x and y')],
    tip="Initialize a scalar field with a function of x and y.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/func_scalar_init.xml')
    )

################

class FuncTwoVectorFieldInit(TwoVectorFieldInit):
    def __init__(self, fx, fy):
        self.fx = fx
        self.fy = fy
    def func(self, position, component):
        if component == 0:
            return self.fx(position.x, position.y)
        return self.fy(position.x, position.y)
    def shortrepr(self):
        return "fx=%s, fy=%s" % (str(self.fx), str(self.fy))


registeredclass.Registration(
    "XYFunction",
    TwoVectorFieldInit,
    FuncTwoVectorFieldInit,
    1,
    params=[
    strfunction.XYStrFunctionParameter('fx', value='0.0',
                   tip="The x component of the field as a function of x and y"),
    strfunction.XYStrFunctionParameter('fy', value='0.0',
                   tip="The y component of the field as a function of x and y"),
    ],
    tip="Initialize a two-vector field with a function of x and y.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/func_twovec_init.xml')
    )
                             
##################

class ScalarOtherMeshInit(OtherMeshFieldInit, ScalarFieldInit):
    pass
    
registeredclass.Registration(
    "Other Mesh",
    ScalarFieldInit,
    ScalarOtherMeshInit,
    2,
    params=[whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                                  tip='Copy field values from this Mesh.')],
    tip="Initialize a scalar field from the same field on another mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/scalar_othermesh_init.xml')
    )

class TwoVectorOtherMeshInit(OtherMeshFieldInit, TwoVectorFieldInit):
    pass
            
registeredclass.Registration(
    "Other Mesh",
    TwoVectorFieldInit,
    TwoVectorOtherMeshInit,
    2,
    params=[whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                                  tip='Copy field values from this Mesh.')],
    tip="Initialize a two-vector field from the same field on another mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/twovec_othermesh_init.xml')
    )

##############################################
    
# Because FieldInitParameters aren't normal RegisteredParameters, the
# classes have to be explicitly added to the xml docs.
xmlmenudump.addRegisteredClass(ScalarFieldInit)
xmlmenudump.addRegisteredClass(TwoVectorFieldInit)

class ResetFieldInit (FieldInit):
    def func(self, position, index):
        return 0
