// -*- C++ -*-
// $RCSfile: material.C,v $
// $Revision: 1.108 $
// $Author: langer $
// $Date: 2005/11/16 22:28:31 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#ifdef DEBUG
#include <assert.h>
#endif
#include "common/activearea.h"
#include "common/cmicrostructure.h"
#include "common/coord.h"
#include "common/pixelattribute.h"
#include "common/IO/stringimage.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "common/printvec.h"
#include "engine/ooferror.h"
#include "engine/property/color/color.h"
#include "engine/IO/propertyoutput.h"
#include "element.h"
#include "femesh.h"
#include "flux.h"
#include "mastercoord.h"
#include "material.h"
#include "equation.h"
#include "timederivative.h"
#include <iostream>
#include <map>

Material::Material(const std::string &nm)
  : name_(nm),
    consistency_state(0),
    fluxprop(Flux::allfluxes().size()),
    fluxfield(Flux::allfluxes().size()),
    outputprop(nPropertyOutputRegistrations())
{}

Material::~Material() {}

bool Material::self_consistent() const {
  return consistency_state.size()==0;
}

const std::vector<std::string> *Material::consistency_messages() const {
  return &consistency_state; 
}

const TimeStamp &Material::getTimeStamp() const {
  return timestamp;
}

std::ostream &operator<<(std::ostream &os, const Material &mat) {
  os << "Material('" << mat.name();
  for(std::vector<Property*>::size_type i=0; i<mat.property.size(); i++)
    os << ", " << mat.property[i];
  os << "')";
  return os;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class MaterialPropertyRegistration {
private:
  const std::string tag;
  Property *property;
  MaterialPropertyRegistration(Property *p, const std::string &str)
    : tag(str), property(p)
  {}
  friend class MaterialPropertyRegistry;
};

// Register a property under the name "tag"
void MaterialPropertyRegistry::registr(Property *prop, const std::string &tag) {
  if(fetch(tag)) {		// Can't have two Properties with the same name
    throw ErrRedundantProperty(tag);
  }
  reg.push_back(new MaterialPropertyRegistration(prop, tag));
}

// Find the property registered with the name "tag"
Property *MaterialPropertyRegistry::fetch(const std::string &tag) const {
  for(std::vector<MaterialPropertyRegistration*>::size_type i=0; i<reg.size();
      i++)
    {
      if(reg[i]->tag == tag) {
	return reg[i]->property;
      }
    }
  return 0;
}

MaterialPropertyRegistry::~MaterialPropertyRegistry() {
  clear();
}

void MaterialPropertyRegistry::clear() {
  for(std::vector<MaterialPropertyRegistration*>::size_type i=0; i<reg.size();
      i++)
    delete reg[i];
  reg.clear();
}

#ifdef DEBUG
void MaterialPropertyRegistry::dump(std::ostream &os) const {
  for(std::vector<MaterialPropertyRegistration*>::size_type i=0; i<reg.size(); i++) {
    os << reg[i]->property->name()
       << " (" << reg[i]->property->classname() << ") "
       << reg[i]->tag << std::endl;
  }
}
#endif // DEBUG

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


Property *Material::getProperty(int i) const { 
  return property[i];
}


void Material::add1Property(Property *prop) {
  property.push_back(prop);
  ++timestamp;
}

void Material::remove1Property(Property *prop) {
  for(std::vector<Property*>::iterator p=property.begin(); p<property.end(); ++p) {
    if(*p == prop) {
      property.erase(p);
      ++timestamp;
      return;
    }
  }
  throw ErrNoSuchProperty(name(), prop->name());
}

void Material::registerPropertyType(Property *p, const std::string &nm) {
  // Called by Property::bookkeeping().
  registry.registr(p, nm);
}

Property *Material::fetchProperty(const std::string &nm) const {
  Property *prop = registry.fetch(nm);
  //  std::cerr << "inside Material::fetchProperty(const string &nm) name=" << nm << std::endl;
  if(!prop)
    throw ErrNoSuchProperty(name(), nm);
  return prop;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Call cross_reference() whenever properties have been added to or
// removed from the material, so the "consistency_state" flag/object is 
// always correct.

void Material::cross_reference() {
  clear_xref();
  consistency_state.clear();  
  

  // ask properties to tell us what kind they are, what fields they use, etc.
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    try {
      // Property::bookkeeping calls:
      //    Material::registerPropertyType (fills in MaterialPropertyRegistry)
      //    Field::registerProperty (which calls Property::require_field)
      //    Material::registerFlux (fills in FluxPropLists)
      property[i]->bookkeeping(this);
    }
    catch(ErrRedundantProperty &exc) {
      // The redundant-property exception is thrown from within the
      // MaterialPropertyRegistry class, when props are registering
      // themselves.  It doesn't know the material -- we could pass it
      // through, but it's easy to just add on here.
      consistency_state.push_back("Property '" + property[i]->name() + \
	"' is a redundant property of type '" + exc.get_tag() + "'.");
    }
  }

  // ask each property to use the registry to find other Properties.
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    try {
      property[i]->cross_reference(this);
    }
    catch(ErrNoSuchProperty &exc) {
      consistency_state.push_back(exc.message());
    }
  }

  // Use FluxPropLists to find which Fields contribute to which Fluxes.
  make_fluxfield_table();
}


void Material::clear_fluxproplist() {
  for(std::vector<FluxPropList>::size_type i=0; i< fluxprop.size(); i++) 
    fluxprop[i].clear();
}

void Material::clear_xref() {
  clear_fluxproplist();
  clear_fluxfield_table();
  clear_outputprop();
  registry.clear();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Material::registerFlux(Property *prop, const Flux *flux) {
  // This property contributes to this flux.  Called by
  // Property::bookkeeping().
//  std::cerr << "Material::registerFlux registering property" << prop->name()<<  " flux=" << flux->name() << std::endl;
  fluxprop[flux->index()].push_back(prop);
}

// Routine to query whether or not this material has properties which
// contribute to the indicated flux -- this is true if the
// corresponding fluxprop entry has nonzero length.
bool Material::contributes_to_flux(const Flux *flux) const {
  if (fluxprop[flux->index()].size() != 0)
    return true;
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Material::registerOutput(Property *prop, const std::string &outputname) {
  // The given Property, which is present in this material,
  // contributes to the given PropertyOutput.
  PropertyOutputRegistration *reg = getPropertyOutputReg(outputname);
  if(!reg)
    throw ErrProgrammingError("PropertyOutputRegistration '" + outputname
			      + "' not found!", __FILE__, __LINE__);
  outputprop[reg->index()].push_back(prop);
}

void Material::clear_outputprop() {
  for(std::vector<std::vector<Property*> >::size_type i=0; i<outputprop.size();
      i++)
    outputprop[i].clear();
};

const std::vector<Property*> &
Material::outputProperties(const PropertyOutput *pout) const {
  return outputprop[pout->index()];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Element start/end.  Since each element has a unique material,
// these can only be called once per element.  The reason they're 
// here is because the Material is the one with the handy-dandy
// list of properties.  
void Material::begin_element(FEMesh *mesh, const Element *el) const {
//   Trace("Material::begin_element");
  for(std::vector<Property*>::size_type i=0;i<property.size();i++) {
    if(property[i]->is_active()) {
      property[i]->begin_element(el);
    }
  }
}

void Material::end_element(FEMesh *mesh, const Element *el) const {
//   Trace("Material::end_element");
  for(std::vector<Property*>::size_type i=0;i<property.size();i++) {
    if(property[i]->is_active()) {
      property[i]->end_element(el);
    }
  }
} 

void Material::precompute(FEMesh *mesh) {
//   std::cerr << "Material::precompute: " << name() << std::endl;
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
//     std::cerr << "Material::precompute: property=" << property[i]->name();
    property[i]->find_active(mesh);
    if(property[i]->is_active()) {
//       std::cerr << " active" << std::endl;
      property[i]->precompute();
    }
//     else
//       std::cerr << " not active" << std::endl;
  }
}

// The cross-referencing relies on properties not being shared
// between materials -- otherwise, the material is ambiguous in 
// precompute(), and in one of the contexts, the property can
// retrieve the wrong companion property.  Each material should
// have a unique copy  of each of its properties, and at this 
// point, a given property instance's precompute() routine should
// be called at most once.


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Post-processing is initially very simple -- just call each
// member property's post_process routine....
void Material::post_process(FEMesh *mesh, const Element *el) const {
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    property[i]->post_process(mesh, el);
  }
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// A Flux is active for this material if any of the Properties that
// contribute to it are active.  A Property is active if all of the
// Fields it uses are defined.

bool FluxPropList::active(const FEMesh *mesh) const {
  for(size_type i=0; i<size(); i++) {
    if((*this)[i]->is_active())
      return true;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Make a list of all fields that contribute to each flux.  Assumes
// that the FluxPropLists have already been constructed by
// registerFlux(), which is called by Property::bookkeeping().  Called
// by Material::cross_reference().

void Material::make_fluxfield_table() {
  // loop over all fluxes
  for(std::vector<FluxPropList>::size_type fluxi=0; fluxi<fluxprop.size();
      fluxi++)
    {
      std::vector<const Field*> &flist = fluxfield[fluxi];

      // loop over all properties that contribute to this flux
      FluxPropList &fpl = fluxprop[fluxi];
      for(FluxPropList::size_type propi=0; propi<fpl.size(); propi++) {
	Property *property = fpl[propi];

	// loop over all fields required by this property
	const std::vector<const Field*> &fields = property->fields_required();
	for(std::vector<const Field*>::size_type fieldi=0; fieldi<fields.size();
	    fieldi++)
	  {
	    const Field *field = fields[fieldi];

	    // this field is needed by this property to compute this flux.
	    bool found = false;	// has it already been listed?
	    for(std::vector<const Field*>::size_type i=0;
		i<flist.size() && !found; i++)
	      {
		if(flist[i] == field) { // yes, it has
		  found = true;
		}
	      }
	    if(!found) {	// no, it hasn't already been listed
	      flist.push_back(field); // list it
	    }
	  }
      }
    }
}

void Material::clear_fluxfield_table() {
  for(std::vector<FluxPropList>::size_type fluxi=0; fluxi<fluxprop.size();
      fluxi++)
    fluxfield[fluxi].clear();
}

// Return a list of Fields relevant to the given Flux in this
// Material.  If the Flux isn't computed by any of this Material's
// Properties, then the list has no entries.

const std::vector<const Field*> &Material::fieldlist(const Flux &flux) const {
  return fluxfield[flux.index()];
}




//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=
//=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=

void Material::make_linear_system(FEMesh *mesh, 
				  const Element *el,
				  StiffnessMatrix *stiffness,
				  StiffnessMatrix *c_matrix,
				  StiffnessMatrix *m_matrix,
				  VECTOR_D &rhs) const
{
  //std::cerr << "Material::make_stiffness" << std::endl;
  // loop over all fluxes
  //   Trace("Material::make_stiffness.");
  for(std::vector<FluxPropList>::size_type fluxi=0; fluxi<fluxprop.size();
      fluxi++)
    {
      Flux *flux = Flux::allfluxes()[fluxi];
      // fluxproplist is the list of Properties that contribute to this Flux
      const FluxPropList &fluxproplist = fluxprop[fluxi];

      // A Flux is active if it is used in one or more active Equations.
      // The fluxproplist is active if any of the Properties that
      // contribute to it are active.  A Property is active if all of
      // the Fields it uses are defined.
      if(mesh->is_active_flux(*flux) && fluxproplist.active(mesh)) {

	// Find the integration order for this flux.  Use the maximum
	// order required by the relevant Properties.  This could be
	// Element dependent, so it can't be precomputed.
	int order = 0;
	for(FluxPropList::size_type propi=0; propi<fluxproplist.size(); propi++)
	  {
	    int ord = fluxproplist[propi]->K_order(mesh, el);
	    if(ord > order) order = ord;
	  }
	order += flux->K_eqn_order(mesh, el);
	
      
	// Integrate over the Element
	for(GaussPointIterator gpt=el->integrator(order); !gpt.end(); ++gpt) {
	  // Trace("Gausspoint " + to_string(gpt.gausspoint()));
	  FluxData *fluxdata = flux->initializeData(el);

	  // Loop over Properties contributing to this Flux,
	  // accumulating their contributions in the flux matrix. First,
	  // get the element's funcnode-iterator and pass it in. Do the
	  // iteration here, though, to ensure that each property in the
	  // fluxproplist gets a crack at each node.

	  for(FluxPropList::size_type propi=0; propi<fluxproplist.size();
	      propi++)
	    {
	      Property *prop = fluxproplist[propi];
	      if(prop->is_active()) {
		//std::cerr << "Material::setting-up flux matrix" << std::endl;
		for(ElementFuncNodeIterator eni = el->funcnode_iterator();
		    !eni.end(); ++eni) {
		  // A flux matrix embodies a numerical representation 
		  // relating the nodal fluxes with its fields.

		  // For elastic systems:
		  // For the in-plane components, it is the "gradient", the 
		  // strain operator, \times the material stiffness,
		  // or equivalently the stiffness \times the nodal derivative 
		  // of the shape function contributions.
		  // For the out of plane components (plane stress), 
		  // it corresponds to the
		  // identity operator, or equivalently, to the stiffness tensor
		  // \times the nodal shape function.
		  // (NOT the derivative) 
	      
		  // For non-elastic systems: 
		  // For in-plane components, it is the "gradient" 
		  // operator \times the material parameter (e.g., thermal
		  // conductivity, mass diffusivity, etc.), or equivalently
		  // a proportionality constant \times the derivative of the
		  // shape function.
		  // For the out of plane components (plane flux) it corresponds
		  // to a proportionality constant times the shape function
		  // (NOT the derivative) 
		  fluxproplist[propi]->fluxmatrix(mesh, el, eni, flux, fluxdata,
						  gpt.gausspoint());
		}
		// For the RHS contributions, don't want to iterate over nodes.
		fluxproplist[propi]->fluxrhs(mesh, el, flux, fluxdata,
					     gpt.gausspoint());
		//fluxrhs embodies the body forces for each driving force
		//for mechanical systems could be gravity \times density,
		//For thermal systems, chemical reactions.
	      } //if "property is active" ends here
	    } //loop over all properties ends here

	  // Have the Flux make its contribution to the stiffness matrix
	  flux->make_stiffness(mesh, el, gpt.gausspoint(), fluxdata,
			       *stiffness);
	  // ...and rhs.
	  flux->make_rhs(mesh, el, gpt.gausspoint(), fluxdata, rhs);

	  delete fluxdata;
	} //loop over gauss points ends here
      } //loop over all fluxproplist ends here
    } //loop over all flux prop ends here
  
  /*
    TIME DERIVATIVES START HERE
  */
  if (mesh->dynamics_activated() || mesh->kinetics_activated())
    set_time_derivatives(mesh, el, c_matrix, m_matrix);
  //time-dependent properties should be inserted at THIS level
  /*
    TIME DERIVATIVES END HERE
  */
}// end of member function



void Material::set_time_derivatives(FEMesh *mesh, 
				    const Element *el,
				    StiffnessMatrix *c_matrix,
				    StiffnessMatrix *m_matrix) const
{
  const std::vector<Equation*>* eqns_list = mesh->all_equations();
  for(std::vector<Equation*>::size_type i=0; i<eqns_list->size(); i++) {
    //for all equations, 	
    Equation * eqn = (*eqns_list)[i];
    const std::vector<TimeDerivative*> timeDervsList=eqn->allTimeDerivatives();
    for(std::vector<TimeDerivative*>::size_type j=0; j<timeDervsList.size();j++)
      {
	TimeDerivative * td = timeDervsList[j];
	// timeDerivproplist is the list of Properties that contribute
	// to this time derivative
	const std::vector<Property*> &timeDerivproplist =
	  td->get_properties(*this); 

	//start loop over all relevant properties and find highest
	//order of integration
	int order = 2*(el->shapefun_degree()); // <Unit> material assumes 2 times shape function degree
	for(std::vector<Property*>::size_type propi=0;
	    propi<timeDerivproplist.size(); propi++)
	  {
	    int ord = timeDerivproplist[propi]->K_order(mesh, el);
	    if(ord > order) 
	      order = ord;
	  } // end loop over all relevant properties and find highest
	    // order of integration
	order += eqn->K_order(el);

	// For every gauss point, every time derivative of every
	// equation contributes to the c_matrix or m_matrix.  Here,
	// c_matrix holds all the first-order time derivatives
	// contributions, and m_matrix embodies all the second order
	// time-derivative contributions from every time-dependent
	// field.

	for(GaussPointIterator gpt=el->integrator(order); !gpt.end(); ++gpt) {
	  // depending on the order of the derivative, each
	  // TimeDerivative object contributes to its C- or M-matrix
	  td->make_matrix(mesh, el, gpt.gausspoint(), c_matrix, m_matrix);
	} //loop over gauss points	
      } //loop over time derivatives of equation
  } //loop over all equations
}


// Find the flux matrix and offset for a single flux at a single point.
void Material::find_fluxmatrix(const FEMesh *mesh, const Element *el,
			       const Flux *flux, FluxData *fluxdata,
			       const MasterPosition &mpos)
  const
{
//   flux->initialize_matrix(el);
  // List of Properties contributing to this flux.
  const FluxPropList &fluxproplist = fluxprop[flux->index()];
  for(FluxPropList::size_type propi=0; propi<fluxproplist.size(); propi++) {
    Property *property = fluxproplist[propi];
    if(property->is_active()) { // are all of the Property's Fields defined?
      for(ElementFuncNodeIterator eni=el->funcnode_iterator();!eni.end();++eni){
	property->fluxmatrix(mesh, el, eni, flux, fluxdata, mpos);
      }
      property->fluxrhs(mesh, el, flux, fluxdata, mpos);
    }
  }
}

//=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=

// Returns a FluxPropList for a give Flux.
FluxPropList Material::get_flux_props(const Flux *fluks) const
{
  return fluxprop[fluks->index()];
}

#ifdef DEBUG
void Material::dump(std::ostream &os) const {
  os << "Material " << name() << ":" << std::endl;
  registry.dump(os);
}

#endif // DEBUG

//=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=

// Materials are assigned to Microstructure pixels with a
// PixelAttribute class.  See common/pixelattribute.h for details.

const std::string
MaterialAttributeRegistration::classname_("MaterialAttributeRegistration");
const std::string
MaterialAttributeRegistration::modulename_("oof2.SWIG.engine.material");

static MaterialAttributeRegistration *matattrreg;

// The MaterialAttributeGlobalData class keeps track of how many
// pixels are assigned to each Material in a Microstructure.  This
// provides a quick way of determining which Materials are used in the
// Microstructure, which allows the Microstructure to respond to
// changes in Materials without looking through all of its pixels to
// see if the changed Material is present.

// Less-than wrapper for materials, used in the std::map.  Required
// because the map's key-value is Material*, not Material.
struct ltMaterial {	
  bool operator()(const Material *m1, const Material *m2) const {
    return m1->name() < m2->name();
  }
};

// MatMap counts how many times each Material is used in the
// Microstructure.
typedef std::map<const Material*, int, ltMaterial> MatMap;

class MaterialAttributeGlobalData : public PixelAttributeGlobalData {
public:
  const CMicrostructure *microstructure;
  MatMap matmap;

  // getCount creats an entry for the material if it does not exist in
  // the map.  If you just want to query, use "count", below.
  int &getCount(const Material *mat) {
    MatMap::iterator i = matmap.find(mat);
    if(i == matmap.end()) {
      matmap[mat] = 0;
      return matmap[mat];
    }
    return (*i).second;
  }

  int count(const Material *mat) const {
    MatMap::const_iterator i = matmap.find(mat);
    if(i == matmap.end()) return 0;
    return (*i).second;
  }

  void remove(const Material *mat) {
    MatMap::iterator i = matmap.find(mat);
    if (i!=matmap.end()) 
      matmap.erase(i);
  }

  MaterialAttributeGlobalData(const CMicrostructure *microstructure)
    : microstructure(microstructure)
  {}
  void materialAddedToPixel(const Material *mat) {
    if(mat)
      getCount(mat)++;
  }
  void materialRemovedFromPixel(const Material *mat) {
    if(mat)
      getCount(mat)--;
  }
//   void materialAddedToAll(const Material *mat) {
//     for(MatMap::iterator i=matmap.begin(); i!=matmap.end(); ++i) {
//       (*i).second = 0;
//     }
//     const ICoord &sz = microstructure->sizeInPixels();
//     getCount(mat) = sz(0)*sz(1);
//   }
//   void clear() {
//     matmap.clear();
//   }

  std::vector<const Material*> *getMaterials() const {
    std::vector<const Material*> *mvec = new std::vector<const Material*>;
    for(MatMap::const_iterator i=matmap.begin(); i!=matmap.end(); ++i) 
      if((*i).second > 0) 
	mvec->push_back((*i).first);
    return mvec;
  }
  TimeStamp getTimeStamp() const {
    TimeStamp latest = timeZero;
    for(MatMap::const_iterator i=matmap.begin(); i!=matmap.end(); ++i) {
      const Material *mat = (*i).first;
      int count = (*i).second;
      if(count > 0) {
	const TimeStamp &ts = mat->getTimeStamp();
	if(ts > latest)
	  latest = ts;
      }
    }
    return latest;
  }
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

MaterialAttributeRegistration::MaterialAttributeRegistration()
  : PxlAttributeRegistration("Material")
{
  matattrreg = this;
}

PixelAttributeGlobalData *
MaterialAttributeRegistration::createAttributeGlobalData(const
							 CMicrostructure *ms)
  const
{
  return new MaterialAttributeGlobalData(ms);
}


void MaterialAttribute::set(const Material *m) {
  material = m;
}

// Find all the MaterialAttributes in the given microstructure which
// refer to this material, and clear them.  If the material count is
// nonzero, iterate over all of of the MaterialAttributes and check
// them individually.

// TODO LATER: It is possible to be more clever about this, having the
// materials maintain a list of attributes which refer to them, and
// keeping it up to date, by tracking all MaterialAttribute set
// operations.  This is currently not done, because it imposes a large
// storage and searching cost on the materials themselves.  Searching
// through the whole array is probably slower, but it happens less
// frequently.

void Material::cleanAttributes(CMicrostructure *ms) const {
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->globalData(ms));
#ifdef DEBUG
  assert(gd!=0);
#endif  
  if (gd->count(this)!=0) {
    const Array<PixelAttribute*> &mat_attrs = matattrreg->map(ms);
    for(Array<PixelAttribute*>::const_iterator i=mat_attrs.begin();
	i!=mat_attrs.end(); ++i) {
      MaterialAttribute *ma = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(ma!=0);
#endif
      if (this == ma->get()) 
	ma->set(0);
    }
    // Remove this material from the map, if it's present.
    gd->remove(this); 
    ms->recategorize();
  }
}

bool MaterialAttribute::operator<(const PixelAttribute &other) const {
  const MaterialAttribute &othermat =
    dynamic_cast<const MaterialAttribute&>(other);
  if(material != 0 && othermat.material != 0)
    return material->name() < othermat.material->name();
  if(material == 0 && othermat.material != 0)
    return true;
  return false;
}

void MaterialAttribute::print(std::ostream &os) const {
  if(material)
    os << "Material(" << material->name() << ")";
  else
    os << "[No Material]";
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Material::assignToPixels(CMicrostructure *microstructure,
			      const std::vector<ICoord> *pxls) const
{
  // First, remove the old material, if any.
  removeMaterialFromPixels(microstructure, *pxls);

  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(std::vector<ICoord>::const_iterator i=pxls->begin(); i<pxls->end(); ++i) {
    if(activearea->isActive(*i)) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(matMap[*i]);
      matAtt->set(this);
      gd->materialAddedToPixel(this);
    }
  }
  microstructure->recategorize();
}

void Material::assignToPixelGroup(CMicrostructure *microstructure,
				  const PixelSet *pixset) const
{
  // Assign material to the pixels in the given group.
  assignToPixels(microstructure, pixset->members());
}

void Material::assignToAllPixels(CMicrostructure *microstructure) const {
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(Array<PixelAttribute*>::iterator i=matMap.begin(); i!=matMap.end(); ++i) {
    if(activearea->isActive(i.coord())) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      gd->materialRemovedFromPixel(matAtt->get());
      matAtt->set(this);
      gd->materialAddedToPixel(this);
    }
  }
  microstructure->recategorize();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

int Material::nPixelsInMicrostructure(const CMicrostructure *microstructure)
  const
{
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  return gd->count(this);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void removeMaterialFromPixels(CMicrostructure *microstructure,
			      const std::vector<ICoord> &pxls)
{
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(std::vector<ICoord>::const_iterator i=pxls.begin(); i<pxls.end(); ++i) {
    if(activearea->isActive(*i)) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(matMap[*i]);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      gd->materialRemovedFromPixel(matAtt->get());
      matAtt->set(0);
    }
  }
  microstructure->recategorize();
}

void removeMaterialFromPixels(CMicrostructure *microstructure,
				const PixelSet *pixset)
{
  removeMaterialFromPixels(microstructure, *pixset->members());
}

void removeAllMaterials(CMicrostructure *microstructure) {
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(Array<PixelAttribute*>::iterator i=matMap.begin(); i!=matMap.end(); ++i) {
    if(activearea->isActive(i.coord())) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      gd->materialRemovedFromPixel(matAtt->get());
      matAtt->set(0);
    }
  }
  microstructure->recategorize();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

TimeStamp getMaterialTimeStamp(const CMicrostructure *microstructure) {
  MaterialAttributeGlobalData *gd =
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  return gd->getTimeStamp();
}

const Material *getMaterialFromCategory(CMicrostructure *microstructure,
					int category)
{
  const ICoord &where = microstructure->getRepresentativePixel(category);
  return getMaterialFromPoint(microstructure, &where);
}

const Material *getMaterialFromPoint(CMicrostructure *microstructure,
				     const ICoord *where)
{
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(matMap[*where]);
#ifdef DEBUG
  assert(matAtt!=0);
#endif
  return matAtt->get();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

StringImage *getMaterialImage(const CMicrostructure *microstructure,
			      const CColor *noMaterial, 
			      const CColor *noColor)
{
  StringImage *image = new StringImage(&microstructure->sizeInPixels(),
				       &microstructure->size());
  const Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  for(Array<PixelAttribute*>::const_iterator i=matMap.begin(); i!=matMap.end();
      ++i)
    {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      const Material *mat = matAtt->get();
      if (mat) {
	try {
	  ColorProp *colorprop =
	    dynamic_cast<ColorProp*>(mat->fetchProperty("Color"));
#ifdef DEBUG
	  assert(colorprop!=0);
#endif
	  const CColor &cc = colorprop->color();
	  image->set(&i.coord(), &cc);
	}
	catch(ErrNoSuchProperty &exc){
	  image->set(&i.coord(), noColor);
	}
      }
      else
	image->set(&i.coord(), noMaterial);
    }
  return image;
}

// Return a list of all Materials used in a Microstructure.
std::vector<const Material*> *
getMaterials(const CMicrostructure *microstructure) {
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  return gd->getMaterials();
}
