// -*- C++ -*-
// $RCSfile: pypropertywrapper.C,v $
// $Revision: 1.28 $
// $Author: langer $
// $Date: 2005/12/13 16:45:21 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */
#include <oofconfig.h>

#include <Python.h>
#include "common/threadstate.h"
#include "common/coord.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/field.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/pypropertywrapper.h"

std::string PyPropertyElementData::classname_("PyPropertyElementData");
std::string PyPropertyElementData::modulename_("oof2.SWIG.engine.pypropertywrapper");

// Note that "regstn" is the registration entry of the property to 
// which the wrapper refers, not that of the wrapper itself.

// The "PythonNative" parent object holds a pointer to the referent
// Python object, and uses it to answer questions about the type of 
// object and so forth.  Because of this, we probably don't have to 
// do indirection on the repr's.
PyPropertyWrapper::PyPropertyWrapper(PyObject *referent, 
				     PyObject *regstn,
				     const std::string &name)
  : PythonNative<Property>(referent), 
    Property(name,regstn),
    referent_(referent)
{ 
  acquirePyLock();
  Py_INCREF(referent_);
  releasePyLock();
}


PyPropertyWrapper::~PyPropertyWrapper() 
{
  acquirePyLock();
  Py_DECREF(referent_);
  releasePyLock();
}

// The routines themselves.  All of these have the same basic
// structure, they are called from the make_stiffness routine
// somewhere, the same as for real C++ properties, and call the
// appropriate Python routine in the referent Python property to get
// the actual answer.


int PyPropertyWrapper::K_order(const FEMesh *mesh, const Element *el) const {
  PyObject *args, *k_result, *func;
  char _element_temp[128];
  char _mesh_temp[128];
  int c_result;
  acquirePyLock();
  try {
    
    func = PyObject_GetAttrString((PyObject *)referent_,"order");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError("PyPropertyWrapper order() not callable.",
				__FILE__,__LINE__);
    }

    SWIG_MakePtr(_mesh_temp, (char*) mesh, "_FEMesh_p");
    SWIG_MakePtr(_element_temp, (char*) el, "_Element_p");
    args = Py_BuildValue("(ss)", _mesh_temp, _element_temp);
    k_result = PyEval_CallObject(func, args);
    Py_XDECREF(args);
    if(k_result==NULL) {
      pythonErrorRelay();
    }

    c_result = PyInt_AsLong(k_result);
    Py_XDECREF(k_result);
  }
  catch (...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
  return c_result;
}

void PyPropertyWrapper::precompute() {
  PyObject *args, *k_result, *func;
  acquirePyLock();
  try {
    func = PyObject_GetAttrString((PyObject *)referent_,"precomputewrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError(
			"PyPropertyWrapper precomputewrap() not callable.",
			__FILE__,__LINE__);
    }
    args = Py_BuildValue("()");
    k_result = PyEval_CallObject(func, args);
    Py_XDECREF(args);
    if(k_result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(k_result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}

void PyPropertyWrapper::cross_reference(Material *mat) {
  PyObject *args, *k_result, *func;
  char _material_temp[128];
  acquirePyLock();
  try {
    func = PyObject_GetAttrString((PyObject *)referent_,"cross_referencewrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError( 
			"PyPropertyWrapper cross_referencewrap() not callable.",
			__FILE__, __LINE__);
    }
    SWIG_MakePtr(_material_temp, (char *)mat, "_Material_p");
    
    args = Py_BuildValue("(s)",_material_temp);

    k_result = PyEval_CallObject(func, args);
    Py_XDECREF(func);

    if(k_result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(k_result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}

void PyPropertyWrapper::begin_element(const Element *el) {
  PyObject *args, *k_result, *func;
  char _element_temp[128];
  acquirePyLock();
  try {
    func = PyObject_GetAttrString((PyObject *)referent_,"begin_elementwrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError(
			"PyPropertyWrapper begin_elementwrap not callable.",
			__FILE__,__LINE__);
    }
    
    SWIG_MakePtr(_element_temp, (char *)el, "_Element_p");
    
    args = Py_BuildValue("(s)",_element_temp);
    k_result = PyEval_CallObject(func, args);
    Py_XDECREF(args);
    if(k_result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(k_result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}

void PyPropertyWrapper::end_element(const Element *el) {
  PyObject *args, *k_result, *func;
  char _element_temp[128];
    
  acquirePyLock();
  try {
    func = PyObject_GetAttrString((PyObject *)referent_,"end_elementwrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError(
			"PyPropertyWrapper end_elementwrap not callable.",
			__FILE__,__LINE__);
    }
    
    SWIG_MakePtr(_element_temp, (char *)el, "_Element_p");
    
    args = Py_BuildValue("(s)",_element_temp);
    k_result = PyEval_CallObject(func, args);
    Py_XDECREF(args);
    if(k_result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(k_result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}


void PyPropertyWrapper::post_process(FEMesh *m, const Element *el)  const {
  char _mesh_temp[128];
  char _element_temp[128];

  PyObject *args,*func,*result;
  acquirePyLock();
  try {
    func = PyObject_GetAttrString((PyObject *)referent_,"post_processwrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError(
			"PyPropertyWrapper post_processwrap not callable.",
			__FILE__, __LINE__);
    }
    SWIG_MakePtr(_mesh_temp, (char *)m, "_FEMesh_p");
    SWIG_MakePtr(_element_temp, (char *)el, "_Element_p");
    args = Py_BuildValue("(ss)", _mesh_temp, _element_temp);
    result = PyEval_CallObject(func, args);
    Py_XDECREF(args);

    if(result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}

// For functions whose arguments are aggregate types, we require
// that the type have a shadow class which is known to SWIG.  Then,
// at this point, we can use the SWIG_MakePtr to create a reference
// to a shadow class instance, and just pass that on to the 
// callback.  
void PyPropertyWrapper::fluxmatrix(const FEMesh *mesh, const Element *el,
				   const ElementFuncNodeIterator &efi,
				   const Flux *flux, FluxData *fluxdata,
				   const MasterPosition &gpt)
  const {
  char _mesh_temp[128];  // SWIG uses 128-byte arrays, also.
  char _element_temp[128];
  char _efni_temp[128];
  char _flux_temp[128];
  char _fluxdata_temp[128];
  char _mp_temp[128];
    
  PyObject *args,*func,*result;
  Trace("PyPropertyWrapper::fluxmatrix");
  
  acquirePyLock();
  try {

    func = PyObject_GetAttrString((PyObject *)referent_, "fluxmatrixwrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError(
			"PyPropertyWrapper fluxmatrixwrap not callable.",
			__FILE__,__LINE__);
    }
    
    SWIG_MakePtr(_mesh_temp, (char *)mesh, "_FEMesh_p");
    SWIG_MakePtr(_element_temp, (char *)el, "_Element_p");
    SWIG_MakePtr(_efni_temp, (char *)(&efi), "_ElementFuncNodeIterator_p");
    SWIG_MakePtr(_flux_temp, (char *)flux, "_Flux_p");
    SWIG_MakePtr(_fluxdata_temp, (char *)fluxdata, "_FluxData_p");
    SWIG_MakePtr(_mp_temp, (char *)(&gpt), "_MasterPosition_p");
    args = Py_BuildValue("(ssssss)",_mesh_temp, _element_temp, _efni_temp,
			 _flux_temp, _fluxdata_temp, _mp_temp);
    result = PyEval_CallObject(func, args);
    Py_XDECREF(args);

    if(result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}


void PyPropertyWrapper::fluxrhs(const FEMesh *mesh, const Element *el,
				const Flux *flux, FluxData *fluxdata,
				const MasterPosition &gpt)
  const {
  char _mesh_temp[128];
  char _element_temp[128];
  char _flux_temp[128];
  char _fluxdata_temp[128];
  char _mp_temp[128];
    
  PyObject *args,*func,*result;

  acquirePyLock();
  try {
    func = PyObject_GetAttrString((PyObject *)referent_, "fluxrhswrap");
    if(!PyCallable_Check(func)) {
      throw ErrProgrammingError(
				"PyPropertyWrapper fluxrhswrap not callable.",
				__FILE__,__LINE__);
    }
    
    SWIG_MakePtr(_mesh_temp, (char *)mesh, "_FEMesh_p");
    SWIG_MakePtr(_element_temp, (char *)el, "_Element_p");
    SWIG_MakePtr(_flux_temp, (char *)flux, "_Flux_p");
    SWIG_MakePtr(_fluxdata_temp, (char *)fluxdata, "_FluxData_p");
    SWIG_MakePtr(_mp_temp, (char *)(&gpt), "_MasterPosition_p");
    args = Py_BuildValue("(sssss)",_mesh_temp, _element_temp,
			 _flux_temp, _fluxdata_temp, _mp_temp);
    result = PyEval_CallObject(func, args);
    Py_XDECREF(args);

    if(result==NULL) {
      pythonErrorRelay();
    }
    Py_XDECREF(result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

PyPropertyElementData::PyPropertyElementData(const std::string & name,
					     PyObject *dat) 
  : ElementData(name), _data(dat) 
{
  acquirePyLock();
  Py_INCREF(_data);
  releasePyLock();
}

// For the particular case of PyObject *'s, simple member 
// retrieval through SWIG doesn't appear to work -- you have to 
// override SWIG's conversion-to-pointer-string to get the actual
// Python object out.  This function has the appropriate typemap
// defined in pypropertywrapper.swg.
PyObject *PyPropertyElementData::data() {
  // Hacky extra incref.
  acquirePyLock();
  Py_INCREF(_data);
  releasePyLock();
  return _data;
}

PyPropertyElementData::~PyPropertyElementData() {
  // It's an error if this refcount is already zero -- avoid XDECREF.
  acquirePyLock();
  Py_DECREF(_data);
  releasePyLock();
}
// PyPropertyElementData-handling functions.  Simple wrappers, mostly.
// These are here partially because Element doesn't have a convenient
// SWIG file, and partially because of the need to do the dynamic
// cast.  More direct Python access to the element functions
// is not out of the question.

// int PyPropertyWrapper::appendElementData(Element *el, 
// 					 PyPropertyElementData *data)
// {
//   return el->appendData(data);
// }

// PyPropertyElementData *PyPropertyWrapper::getElementData(Element *el, int i)
// {
//   return dynamic_cast<PyPropertyElementData*>(el->getData(i));
// }

// int PyPropertyWrapper::getElementDataIndex(Element *el, std::string name)
// {
//   std::cerr << "getElementDataIndex" << std::endl;
//   return el->getIndexByName(name);
// }

